import re

from unidecode import unidecode

CITY_TO_PROVINCE = {"Madrid": "Madrid"}

POSTCODES = {
    "01": "Álava",
    "02": "Albacete",
    "03": "Alicante",
    "04": "Almería",
    "05": "Ávila",
    "06": "Badajoz",
    "07": "Baleares",
    "08": "Barcelona",
    "09": "Burgos",
    "10": "Cáceres",
    "11": "Cádiz",
    "12": "Castellón",
    "13": "Ciudad Real",
    "14": "Córdoba",
    "15": "La Coruña",
    "16": "Cuenca",
    "17": "Gerona",
    "18": "Granada",
    "19": "Guadalajara",
    "20": "Guipúzcoa",
    "21": "Huelva",
    "22": "Huesca",
    "23": "Jaén",
    "24": "León",
    "25": "Lérida",
    "26": "La Rioja",
    "27": "Lugo",
    "28": "Madrid",
    "29": "Málaga",
    "30": "Murcia",
    "31": "Navarra",
    "32": "Orense",
    "33": "Asturias",
    "34": "Palencia",
    "35": "Las Palmas",
    "36": "Pontevedra",
    "37": "Salamanca",
    "38": "Santa Cruz de Tenerife",
    "39": "Cantabria",
    "40": "Segovia",
    "41": "Sevilla",
    "42": "Soria",
    "43": "Tarragona",
    "44": "Teruel",
    "45": "Toledo",
    "46": "Valencia",
    "47": "Valladolid",
    "48": "Vizcaya",
    "49": "Zamora",
    "50": "Zaragoza",
    "51": "Ceuta",
    "52": "Melilla",
}

# Mapping of normalized names (for comparison) to standardized names (for storing)
# for each corresponding country code
STATES = {
    "es": {
        "andalucia": "Andalucía",
        "aragon": "Aragón",
        "asturias": "Asturias",
        "baleares": "Baleares",
        "canarias": "Canarias",
        "cantabria": "Cantabria",
        "castilla la mancha": "Castilla-La Mancha",
        "castilla y leon": "Castilla y León",
        "cataluna": "Cataluña",
        "ceuta": "Ceuta",
        "comunidad valenciana": "Comunidad Valenciana",
        "extremadura": "Extremadura",
        "galicia": "Galicia",
        "la rioja": "La Rioja",
        "madrid": "Comunidad de Madrid",
        "melilla": "Melilla",
        "murcia": "Murcia",
        "navarra": "Navarra",
        "pais vasco": "País Vasco",
        "euskadi": "País Vasco",  # Alias not caught by rapidfuzz
    }
}

PROVINCES = {
    "es": {
        "alava": "Álava",
        "araba": "Álava",  # Alias not caught by rapidfuzz
        "albacete": "Albacete",
        "alicante": "Alicante",
        "almeria": "Almería",
        "asturias": "Asturias",
        "avila": "Ávila",
        "badajoz": "Badajoz",
        "barcelona": "Barcelona",
        "bizkaia": "Vizcaya",
        "burgos": "Burgos",
        "caceres": "Cáceres",
        "cadiz": "Cádiz",
        "cantabria": "Cantabria",
        "castellon": "Castellón",
        "ceuta": "Ceuta",  # Considered province by opensm and/or geonames
        "ciudad real": "Ciudad Real",
        "cordoba": "Córdoba",
        "cuenca": "Cuenca",
        "gipuzkoa": "Gipuzkoa",
        "gerona": "Gerona",
        "granada": "Granada",
        "guadalajara": "Guadalajara",
        "huelva": "Huelva",
        "huesca": "Huesca",
        "islas baleares": "Islas Baleares",
        "jaen": "Jaén",
        "la coruna": "La Coruña",
        "la rioja": "La Rioja",
        "las palmas": "Las Palmas",
        "leon": "León",
        "lerida": "Lérida",
        "lugo": "Lugo",
        "madrid": "Madrid",
        "malaga": "Málaga",
        "melilla": "Melilla",  # Considered province by opensm and/or geonames
        "murcia": "Murcia",
        "navarra": "Navarra",
        "orense": "Orense",
        "palencia": "Palencia",
        "pontevedra": "Pontevedra",
        "salamanca": "Salamanca",
        "santa cruz de tenerife": "Santa Cruz de Tenerife",
        "segovia": "Segovia",
        "sevilla": "Sevilla",
        "soria": "Soria",
        "tarragona": "Tarragona",
        "teruel": "Teruel",
        "toledo": "Toledo",
        "valencia": "Valencia",
        "valladolid": "Valladolid",
        "zamora": "Zamora",
        "zaragoza": "Zaragoza",
    }
}


PROVINCE_TO_POSTCODE = {
    "es": {
        "A Coruña": "15",
        "Álava": "01",
        "Araba": "01",
        "Alacant": "03",
        "Alicante": "03",
        "Albacete": "02",
        "Almería": "04",
        "Asturias": "33",
        "Ávila": "05",
        "Badajoz": "06",
        "Baleares": "07",
        "Barcelona": "08",
        "Bizkaia": "48",
        "Burgos": "09",
        "Cáceres": "10",
        "Cádiz": "11",
        "Cantabria": "39",
        "Castelló": "12",
        "Castellón": "12",
        "Ceuta": "51",
        "Ciudad Real": "13",
        "Córdoba": "14",
        "Cuenca": "16",
        "Gerona": "17",
        "Gipuzkoa": "20",
        "Girona": "17",
        "Granada": "18",
        "Guadalajara": "19",
        "Guipúzcoa": "20",
        "Huelva": "21",
        "Huesca": "22",
        "Illes Balears": "07",
        "Jaén": "23",
        "La Coruña": "15",
        "La Rioja": "26",
        "Las Palmas": "35",
        "León": "24",
        "Lérida": "25",
        "Lleida": "25",
        "Lugo": "27",
        "Madrid": "28",
        "Málaga": "29",
        "Melilla": "52",
        "Murcia": "30",
        "Navarra": "31",
        "Orense": "32",
        "Ourense": "32",
        "Palencia": "34",
        "Pontevedra": "36",
        "Salamanca": "37",
        "Santa Cruz de Tenerife": "38",
        "Segovia": "40",
        "Sevilla": "41",
        "Soria": "42",
        "Tarragona": "43",
        "Teruel": "44",
        "Toledo": "45",
        "València": "46",
        "Valencia": "46",
        "Valladolid": "47",
        "Vizcaya": "48",
        "Zamora": "49",
        "Zaragoza": "50",
    },
    "pt": {
        "Aveiro": "3",
        "Beja": "7",
        "Braga": "4",
        "Bragança": "5",
        "Castelo Branco": "6",
        "Coimbra": "3",
        "Évora": "7",
        "Faro": "8",
        "Guarda": "6",
        "Leiria": "2",
        "Lisboa": "1",
        "Portalegre": "7",
        "Porto": "4",
        "Santarém": "2",
        "Setúbal": "2",
        "Viana do Castelo": "4",
        "Vila Real": "5",
        "Viseu": "3",
        "Açores": "9",
        "Madeira": "9",
    },
}


POSTCODE_TO_STATES = {
    "es": {
        # Andalucía
        "04": "Andalucía",
        "11": "Andalucía",
        "14": "Andalucía",
        "18": "Andalucía",
        "21": "Andalucía",
        "23": "Andalucía",
        "29": "Andalucía",
        "41": "Andalucía",
        # Aragón
        "22": "Aragón",
        "44": "Aragón",
        "50": "Aragón",
        # Asturias
        "33": "Principado de Asturias",
        # Baleares
        "07": "Islas Baleares",
        # Canarias
        "35": "Canarias",
        "38": "Canarias",
        # Cantabria
        "39": "Cantabria",
        # Castilla y León
        "05": "Castilla y León",
        "09": "Castilla y León",
        "24": "Castilla y León",
        "34": "Castilla y León",
        "37": "Castilla y León",
        "40": "Castilla y León",
        "42": "Castilla y León",
        "47": "Castilla y León",
        "49": "Castilla y León",
        # Castilla-La Mancha
        "02": "Castilla-La Mancha",
        "13": "Castilla-La Mancha",
        "16": "Castilla-La Mancha",
        "19": "Castilla-La Mancha",
        "45": "Castilla-La Mancha",
        # Cataluña
        "08": "Cataluña",
        "17": "Cataluña",
        "25": "Cataluña",
        "43": "Cataluña",
        # Comunidad Valenciana
        "03": "Comunidad Valenciana",
        "12": "Comunidad Valenciana",
        "46": "Comunidad Valenciana",
        # Extremadura
        "06": "Extremadura",
        "10": "Extremadura",
        # Galicia
        "15": "Galicia",
        "27": "Galicia",
        "32": "Galicia",
        "36": "Galicia",
        # Madrid
        "28": "Comunidad de Madrid",
        # Murcia
        "30": "Región de Murcia",
        # Navarra
        "31": "Comunidad Foral de Navarra",
        # País Vasco
        "01": "País Vasco",
        "20": "País Vasco",
        "48": "País Vasco",
        # La Rioja
        "26": "La Rioja",
        # Ciudades Autónomas
        "51": "Ceuta",
        "52": "Melilla",
    },
    "pt": {  # --- NORTE ---
        "40": "Porto",
        "41": "Porto",
        "42": "Porto",
        "43": "Porto",
        "44": "Porto",
        "45": "Aveiro",  # Concelhos do norte de Aveiro, na fronteira com Porto.
        "47": "Braga",
        "48": "Braga",  # Guimarães.
        "49": "Viana do Castelo",
        "50": "Vila Real",
        "51": "Vila Real",
        "52": "Vila Real",
        "53": "Vila Real / Bragança",  # Zona fronteiriça.
        "54": "Bragança",
        # --- CENTRO ---
        "60": "Castelo Branco",
        "61": "Castelo Branco",
        "62": "Castelo Branco",
        "63": "Guarda",
        "30": "Coimbra",
        "31": "Coimbra",
        "32": "Coimbra",
        "33": "Coimbra",
        "34": "Viseu",
        "35": "Viseu",
        "37": "Aveiro",
        "38": "Aveiro",
        "24": "Leiria",
        # --- ÁREA METROPOLITANA DE LISBOA e arredores ---
        "10": "Lisboa",
        "11": "Lisboa",
        "12": "Lisboa",
        "13": "Lisboa",
        "14": "Lisboa",
        "15": "Lisboa",
        "16": "Lisboa",
        "17": "Lisboa",
        "18": "Lisboa",
        "19": "Lisboa",
        "20": "Santarém",
        "21": "Santarém",
        "22": "Santarém",
        "23": "Santarém",  # Tomar e Torres Novas.
        "25": "Lisboa",  # Concelhos como Torres Vedras, Mafra, Alenquer.
        "26": "Lisboa",  # Concelhos como Loures, Amadora, Odivelas.
        "27": "Lisboa",  # Concelhos como Sintra, Cascais, Oeiras.
        "28": "Setúbal",
        "29": "Setúbal",
        # --- ALENTEJO ---
        "70": "Évora",
        "71": "Évora",
        "72": "Évora",
        "73": "Portalegre",
        "74": "Portalegre",
        "75": "Setúbal",  # Litoral Alentejano (Sines, Grândola), administrativamente de Setúbal.
        "76": "Beja",
        "77": "Beja",
        "78": "Beja",
        "79": "Beja",
        # --- ALGARVE ---
        "80": "Faro",
        "81": "Faro",
        "82": "Faro",
        "83": "Faro",
        "84": "Faro",
        "85": "Faro",
        "86": "Faro",
        "87": "Faro",
        "88": "Faro",
        "89": "Faro",
        # --- REGIÕES AUTÓNOMAS ---
        "90": "Madeira",
        "91": "Madeira",
        "92": "Madeira",
        "93": "Madeira",
        "95": "Açores",  # Ilha de São Miguel (Ponta Delgada).
        "96": "Açores",  # Ilha de São Miguel (Ribeira Grande) e Santa Maria.
        "97": "Açores",  # Ilha Terceira (Angra do Heroísmo).
        "98": "Açores",  # Ilhas de São Jorge, Graciosa, Faial, Pico.
        "99": "Açores",  # Ilhas de Flores e Corvo.
    },
}

_NORMALIZED_PROVINCE_CACHE = {}
for country, provinces in PROVINCE_TO_POSTCODE.items():
    # Get the original keys (e.g., "A Coruña", "Álava")
    original_keys = list(provinces.keys())

    # Create the normalized list (e.g., "a coruna", "alava")
    normalized_choices = [unidecode(p).lower() for p in original_keys]

    _NORMALIZED_PROVINCE_CACHE[country] = {
        "choices": normalized_choices,  # The list for rapidfuzz to search in
        "keys": original_keys,  # The list to find the name by index
    }

# Source: https://github.com/ariankoochak/regex-patterns-of-all-countries
COUNTRY_PARSING_RULES = {
    "es": {
        "zip_validate_pattern": re.compile(r"^\d{5}$"),
        "zip_search_pattern": re.compile(r"\b\d{5}\b"),
        "phone_validate_pattern": re.compile(r"^(\+?34)?[67]\d{8}$"),
    },
    "pt": {
        "zip_validate_pattern": re.compile(r"^\d{4}[- ]{0,1}\d{3}$|^\d{4}$"),
        "zip_search_pattern": re.compile(r"\b\d{4}[- ]?\d{3}\b|\b\d{4}\b"),
        "phone_validate_pattern": re.compile(r"^(\+?351)?9[1236]\d{7}$"),
    },
}

# Cutoff score for rapidfuzz in the name standardization function
STANDARD_THRESHOLD = 40



########################################################################################################################
# IMPORTS

import logging
import random
import time
from datetime import timedelta
from functools import partial

import requests
import tenacity
from stem import Signal
from stem.control import Controller

from datamarket.exceptions import EnsureNewIPTimeoutError, NoWorkingProxiesError

########################################################################################################################
# SETUP

logger = logging.getLogger(__name__)
logging.getLogger("stem").setLevel(logging.WARNING)

PROXY_ROTATION_INTERVAL = timedelta(minutes=10)
PROXY_ROTATION_TIMEOUT_SECONDS = int(PROXY_ROTATION_INTERVAL.total_seconds())

########################################################################################################################
# CLASSES


class ProxyInterface:
    """
    Manage HTTP, HTTPS, and SOCKS5 proxies configured in the [proxy] section.
    """

    CHECK_IP_URL = "https://wtfismyip.com/json"

    def __init__(self, config):
        self._load_from_config(config)
        self.current_index = -2  # -2 means no selection made yet, -1 means Tor selected
        self._health = {}  # {entry: {"ok": bool, "last_checked": time.time(), "last_error": str}}
        self._traversal_queue = []  # Queue of indices left to test in current traversal
        self._traversal_start = None  # Timestamp when current traversal started
        self._last_ip_wait = {}  # {entry: (timestamp, traversal_cycle)} - last time we attempted to wait for IP change
        self._traversal_cycle = 0  # Counter of full traversals of the queue

    def _load_from_config(self, cfg):
        # Tor password (optional)
        self.tor_password = cfg.get("proxy", "tor_password", fallback=None)

        # Comma-separated list of hosts
        hosts_raw = cfg.get("proxy", "hosts", fallback="")
        if not hosts_raw:
            raise RuntimeError("[proxy] hosts list is empty")

        entries = []
        for host_entry in (h.strip() for h in hosts_raw.split(",") if h.strip()):
            host, port, user, password = self._parse_host_entry(host_entry)
            entries.append((host, port, user, password))

        self.entries = entries

    def _parse_host_entry(self, host_entry):
        if "@" in host_entry:
            auth_part, host_part = host_entry.rsplit("@", 1)
            host, port = host_part.split(":")
            user, password = auth_part.split(":", 1)
            return host, port, user, password
        else:
            host, port = host_entry.split(":")
            return host, port, None, None

    @property
    def proxies(self):
        return self.get_proxies(use_tor=bool(self.tor_password))

    @staticmethod
    def get_proxy_url(host, port, user=None, password=None, schema="http"):
        auth = f"{user}:{password}@" if user and password else ""
        return f"{schema}://{auth}{host}:{port}"

    def get_proxies(
        self,
        use_tor=False,
        randomize=False,
        raw=False,
        use_auth=False,
        use_socks=False,
        check_timeout=5,
        cooldown_seconds=30,
        proxy_rotation_interval=PROXY_ROTATION_INTERVAL,
    ):
        """
        Return parsed proxy URLs or raw entry tuple for a working proxy.

        :param use_tor: route via local Tor SOCKS5 if True
        :param randomize: select a random proxy if True, otherwise round-robin
        :param raw: return raw (host, port, user, password) tuple if True
        :param use_auth: include proxies that require authentication if True; otherwise only credential-free
        :param check_timeout: timeout in seconds for health check requests
        :param cooldown_seconds: how long to cache health status before re-checking
        :param proxy_rotation_interval: max time to retry finding working proxies (timedelta or seconds, 0 to disable)
        """
        # Tor handling (skip health check for tor)
        if use_tor:
            self.current_index = -1  # Indicate Tor is selected
            if raw:
                return ("127.0.0.1", "9050", None, None)
            return {"socks5": self.get_proxy_url("127.0.0.1", 9050, schema="socks5")}

        # Get a working entry (with health checks if enabled)
        host, port, user, password = self._get_working_entry(
            use_auth=use_auth,
            randomize=randomize,
            check_timeout=check_timeout,
            cooldown_seconds=cooldown_seconds,
            proxy_rotation_interval=proxy_rotation_interval,
        )

        if raw:
            return host, port, user, password

        # Build mapping of proxy URLs
        if use_socks:
            return {
                "socks5": self.get_proxy_url(host, port, user, password, "socks5"),
            }
        else:
            return {
                "http": self.get_proxy_url(host, port, user, password, "http"),
                "https": self.get_proxy_url(host, port, user, password, "http"),
            }

    def check_current_ip(self, proxies=None):
        try:
            proxies_arg = proxies or {"http": self.proxies["http"]}
            resp = requests.get(self.CHECK_IP_URL, proxies=proxies_arg, timeout=30)
            return resp.json().get("YourFuckingIPAddress")
        except Exception as ex:
            logger.error(ex)

    def renew_tor_ip(self):
        if not self.tor_password:
            logger.error("Tor password not configured")
            return

        try:
            logger.debug(f"Current IP: {self.check_current_ip()}")
            with Controller.from_port(port=9051) as controller:
                controller.authenticate(password=self.tor_password)
                controller.signal(Signal.NEWNYM)

            time.sleep(5)
            logger.debug(f"New IP: {self.check_current_ip()}")
        except Exception as ex:
            logger.error("Failed to renew Tor IP")
            logger.error(ex)

    def wait_for_new_ip(self, timeout=PROXY_ROTATION_TIMEOUT_SECONDS, interval=30, check_timeout=5):
        """
        Ensures that the IP address of the selected proxy differs from any other proxy chosen within the proxy rotation interval.

        :param timeout: Max seconds to wait for IP change
        :param interval: Seconds between IP checks
        :param check_timeout: Timeout for individual IP check requests
        :return: The selected entry (unchanged)
        :raises RuntimeError: If no proxy is available or baseline cannot be determined
        :raises EnsureNewIPTimeoutError: If IP doesn't change within timeout
        """
        # Use currently selected proxy
        if self.current_index == -1:
            # Tor is selected
            entry = ("127.0.0.1", "9050", None, None)
        elif self.current_index >= 0 and self.current_index < len(self.entries):
            # current_index points to the selected entry
            entry = self.entries[self.current_index]
        else:
            # No valid selection, select one
            logger.debug("No proxy currently selected, selecting one for IP waiting")
            self.get_proxies(raw=True)
            if self.current_index == -1:
                entry = ("127.0.0.1", "9050", None, None)
            elif self.current_index >= 0 and self.current_index < len(self.entries):
                entry = self.entries[self.current_index]
            else:
                raise RuntimeError("Could not select a proxy for IP waiting")

        # Check if we should skip waiting based on global cooldown and traversal cycle
        now = time.time()
        interval_seconds = PROXY_ROTATION_INTERVAL.total_seconds()
        for last_wait in self._last_ip_wait.values():
            last_ts, last_cycle = last_wait

            time_recent = last_ts is not None and (now - last_ts) <= interval_seconds
            no_full_rotation = self._traversal_cycle <= last_cycle

            # Skip only if both conditions are true: recent wait AND no full traversal cycle since
            if time_recent and no_full_rotation:
                logger.debug(
                    "Skipping wait_for_new_ip: last wait %.1fs ago and no full traversal since (last_cycle=%s current=%s)",
                    now - last_ts,
                    last_cycle,
                    self._traversal_cycle,
                )
                return

        # Mark we are now attempting to wait for this entry
        self._last_ip_wait[entry] = (now, self._traversal_cycle)

        # Try to use cached baseline IP from health check
        health = self._health.get(entry, {})
        baseline = health.get("last_ip")
        if baseline is None:
            raise RuntimeError(f"Could not determine baseline IP for entry {entry[0]}:{entry[1]}")
        return self._wait_for_new_ip(entry, baseline, timeout, interval, check_timeout)

    def _mark_entry_status(self, entry, ok, error=None, last_ip=None):
        """Update health cache for an entry."""
        self._health[entry] = {
            "ok": ok,
            "last_checked": time.time(),
            "last_error": error,
            "last_ip": last_ip,
        }

    def _is_entry_alive(self, entry, timeout=5):
        """Check if a proxy entry is working by making a test request."""
        host, port, user, pwd = entry
        try:
            proxies = {
                "http": self.get_proxy_url(host, port, user, pwd, "http"),
                "https": self.get_proxy_url(host, port, user, pwd, "http"),
            }
            resp = requests.get(self.CHECK_IP_URL, proxies=proxies, timeout=timeout)
            ok = resp.status_code == 200
            last_ip = resp.json().get("YourFuckingIPAddress") if ok else None
            self._mark_entry_status(entry, ok, last_ip=last_ip)
            return ok
        except Exception as ex:
            self._mark_entry_status(entry, False, str(ex))
            return False

    def _get_working_entry(
        self,
        use_auth=False,
        randomize=False,
        check_timeout=5,
        cooldown_seconds=30,
        proxy_rotation_interval=PROXY_ROTATION_INTERVAL,
    ):
        """Get a working proxy entry, performing health checks as needed.

        - Fails fast if there are no entries.
        - Optionally retries for up to `proxy_rotation_interval`,
        refreshing the traversal queue before each attempt.
        """

        if not self.entries:
            raise NoWorkingProxiesError("No proxies available")

        pool = self._build_pool(use_auth)
        self._refresh_traversal_queue(pool, randomize)

        find_once = partial(self._find_working_entry_once, check_timeout, cooldown_seconds)

        if not proxy_rotation_interval:
            return find_once()

        def before_sleep(retry_state):
            tenacity.before_sleep_log(logger, logging.INFO)(retry_state)
            self._refresh_traversal_queue(pool, randomize)

        retrying = tenacity.Retrying(
            wait=tenacity.wait_fixed(cooldown_seconds),
            stop=tenacity.stop_after_delay(proxy_rotation_interval),
            before_sleep=before_sleep,
            retry=tenacity.retry_if_exception_type(NoWorkingProxiesError),
            reraise=True,
        )
        return retrying(find_once)

    def _get_round_robin_candidates(self, pool):
        """Get candidates in round-robin order starting from current_index."""
        candidates = []
        start_idx = (self.current_index + 1) % len(self.entries) if self.current_index >= 0 else 0
        for i in range(len(self.entries)):
            idx = (start_idx + i) % len(self.entries)
            entry = self.entries[idx]
            if entry in pool:
                candidates.append(idx)
        return candidates

    def _build_pool(self, use_auth):
        pool = self.entries if use_auth else [e for e in self.entries if not e[2] and not e[3]]
        if not pool:
            pool = self.entries
        return pool

    def _refresh_traversal_queue(self, pool, randomize):
        # Build current pool indices
        current_pool_indices = [idx for idx, entry in enumerate(self.entries) if entry in pool]

        # Check if we need to refill the traversal queue
        if not self._traversal_queue and current_pool_indices:
            if randomize:
                self._traversal_queue = current_pool_indices.copy()
                random.shuffle(self._traversal_queue)
            else:
                # Round-robin: start from next after current_index
                self._traversal_queue = []
                start_idx = (self.current_index + 1) % len(self.entries) if self.current_index >= 0 else 0
                for i in range(len(self.entries)):
                    idx = (start_idx + i) % len(self.entries)
                    if idx in current_pool_indices:
                        self._traversal_queue.append(idx)
            self._traversal_start = time.time()
            self._traversal_cycle += 1

    def _find_working_entry_once(self, check_timeout, cooldown_seconds):
        # Consume from traversal queue for cached checks
        for idx in self._traversal_queue:
            entry = self.entries[idx]
            health = self._health.get(entry, {})
            last_checked = health.get("last_checked", 0)
            ok = health.get("ok", False)
            now = time.time()

            if ok and (now - last_checked) < cooldown_seconds:
                logger.debug(f"Using cached working proxy: {entry[0]}:{entry[1]}")
                self.current_index = idx
                self._traversal_queue.remove(idx)
                return entry
            elif not ok and (now - last_checked) < cooldown_seconds:
                continue
            else:
                logger.debug(f"Checking proxy health: {entry[0]}:{entry[1]}")
                if self._is_entry_alive(entry, timeout=check_timeout):
                    self.current_index = idx
                    self._traversal_queue.remove(idx)
                    return entry

        self._traversal_queue = []
        raise NoWorkingProxiesError("No working proxies available")

    def _wait_for_new_ip(self, entry, baseline, timeout, interval, check_timeout):
        logger.info("Refreshing proxy IP...")
        start = time.time()
        while time.time() - start < timeout:
            host, port, user, pwd = entry
            proxies_map = {
                "http": self.get_proxy_url(host, port, user, pwd, "http"),
                "https": self.get_proxy_url(host, port, user, pwd, "http"),
            }
            try:
                resp = requests.get(self.CHECK_IP_URL, proxies=proxies_map, timeout=check_timeout)
                current_ip = resp.json().get("YourFuckingIPAddress")
            except Exception:
                current_ip = None

            if current_ip and current_ip != baseline:
                self._mark_entry_status(entry, True, last_ip=current_ip)
                logger.info(f"IP changed from {baseline} to {current_ip}")
                return

            time.sleep(interval)

        raise EnsureNewIPTimeoutError(f"Timed out waiting for new IP after {timeout}s")



########################################################################################################################
# IMPORTS

import io
import logging
from typing import Any, Dict, List, Optional

import boto3

########################################################################################################################
# CLASSES

logger = logging.getLogger(__name__)


class AWSInterface:
    def __init__(self, config) -> None:
        self.profiles: List[Dict[str, Any]] = []
        self.config = config

        for section in getattr(self.config, "sections", lambda: [])():
            if section.startswith("aws:"):
                profile_name = section.split(":", 1)[1]
                bucket_value = self.config[section].get("buckets", "")
                buckets = [b.strip() for b in bucket_value.split(",") if b.strip()]
                session = boto3.Session(profile_name=profile_name)

                self.profiles.append(
                    {
                        "profile": profile_name,
                        "buckets": buckets,
                        "session": session,
                    }
                )

        if not self.profiles:
            logger.warning("No AWS profiles found in config file")

        self.current_profile: Optional[Dict[str, Any]] = self.profiles[0] if self.profiles else None
        self._update_resources()

    def _update_resources(self) -> None:
        """Refresh S3 resources for the current profile and set default bucket (first in list)"""
        if self.current_profile:
            self.s3 = self.current_profile["session"].resource("s3")
            self.s3_client = self.s3.meta.client
            buckets = self.current_profile.get("buckets", [])
            self.bucket = buckets[0] if buckets else None
        else:
            self.s3 = None
            self.s3_client = None
            self.bucket = None

    def switch_profile(self, profile_name: str) -> None:
        for profile in self.profiles:
            if profile["profile"] == profile_name:
                self.current_profile = profile
                self._update_resources()
                return
        logger.warning(f"Profile {profile_name} not found")

    def switch_bucket(self, bucket: str) -> None:
        if not self.current_profile:
            logger.warning("No current AWS profile to switch bucket on")
            return

        buckets = self.current_profile.get("buckets") or []
        if bucket not in buckets:
            logger.warning(f"Bucket {bucket} not found in profile {self.current_profile.get('profile')}")
            return

        self.bucket = bucket

    def switch_bucket_for_profile(self, profile_name: str, bucket: str) -> None:
        """
        Select a profile and then switch its active bucket.
        """
        for profile in self.profiles:
            if profile["profile"] == profile_name:
                self.current_profile = profile
                self._update_resources()  # sets default bucket & s3 clients
                self.switch_bucket(bucket)  # only sets self.bucket if valid
                return
        logger.warning(f"Profile {profile_name} not found")

    def get_bucket_url(self) -> Optional[str]:
        """Return active bucket URL."""
        if not self.bucket:
            logger.warning("No active bucket selected")
            return None
        region = self.s3_client.meta.region_name
        return f"https://{self.bucket}.s3.{region}.amazonaws.com"

    def get_file(self, s3_path: str):
        if not self.bucket:
            logger.warning("No active bucket selected")
            return None
        try:
            return self.s3.Object(self.bucket, s3_path).get()
        except self.s3_client.exceptions.NoSuchKey:
            logger.info(f"{s3_path} does not exist")
            return None

    def file_exists(self, s3_path: str) -> bool:
        if not self.bucket:
            logger.warning("No active bucket selected")
            return False
        try:
            self.s3_client.head_object(Bucket=self.bucket, Key=s3_path)
            return True
        except self.s3_client.exceptions.NoSuchKey:
            return False
        except Exception as e:
            logger.error(f"Error checking existence of {s3_path}: {e}")
            raise

    def read_file_as_bytes(self, s3_path: str) -> Optional[io.BytesIO]:
        obj = self.get_file(s3_path)
        if not obj:
            return None
        return io.BytesIO(obj["Body"].read())

    def upload_file(self, local_path: str, s3_path: str, **kwargs) -> None:
        if not self.bucket:
            logger.warning("No active bucket selected")
            return
        self.s3.Bucket(self.bucket).upload_file(local_path, s3_path, **kwargs)



########################################################################################################################
# IMPORTS

import logging
from datetime import timedelta
from typing import Optional, Sequence

import requests
from bs4 import BeautifulSoup
from requests.exceptions import ProxyError
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    retry_if_not_exception_type,
    stop_after_attempt,
    stop_after_delay,
    wait_exponential,
)

from datamarket.exceptions.main import IgnoredHTTPError

from ..exceptions import BadRequestError, EmptyResponseError, NotFoundError, RedirectionDetectedError
from ..interfaces.proxy import ProxyInterface
from .main import ban_sleep

########################################################################################################################
# SETUP LOGGER

logger = logging.getLogger(__name__)


########################################################################################################################
# CLASSES


class RequestsClient:
    """A robust, proxy-enabled HTTP client with retry logic and flexible output formats."""

    def __init__(self, proxy_interface: Optional[ProxyInterface] = None):
        """
        Initializes the HTTP client.

        Args:
            proxy_interface (Optional[ProxyInterface], optional): Proxy provider. If None, no proxy is used. Defaults to None.
        """
        self.proxy_interface = proxy_interface

    @retry(
        retry=retry_if_exception_type(ProxyError),
        wait=wait_exponential(exp_base=3, multiplier=3, max=60),
        stop=stop_after_delay(timedelta(minutes=10)),
        before_sleep=before_sleep_log(logger, logging.WARNING),
        reraise=True,
    )
    def _request_with_proxy_retry(self, url: str, method: str, use_auth: bool, **params):
        """
        Performs an HTTP request with proxy retry logic.

        Args:
            url (str): The target URL to fetch data from.
            method (str): HTTP method to use (e.g., 'GET', 'POST').
            use_auth (bool): Whether to use authenticated proxies.
            **params: Additional arguments passed to the requests method.

        Returns:
            requests.Response: The HTTP response object.
        """
        logger.info(f"Fetching data from {url} ...")
        proxy_cfg = None
        if self.proxy_interface:
            host, port, user, pwd = self.proxy_interface.get_proxies(raw=True, use_auth=use_auth)
            if host and port:
                proxy_url = f"http://{host}:{port}"
                proxy_auth_url = f"http://{user}:{pwd}@{host}:{port}"
                proxy_cfg = {"http": proxy_url, "https": proxy_url}
                if user and pwd:
                    proxy_cfg = {"http": proxy_auth_url, "https": proxy_auth_url}
                logger.info(f"Using proxy: {proxy_url}")
        response = getattr(requests, method.lower())(url, proxies=proxy_cfg, **params)

        return response

    @retry(
        retry=retry_if_not_exception_type(
            (IgnoredHTTPError, NotFoundError, BadRequestError, RedirectionDetectedError, ProxyError)
        ),
        wait=wait_exponential(exp_base=3, multiplier=3, max=60),
        stop=stop_after_attempt(5),
        before_sleep=before_sleep_log(logger, logging.WARNING),
        reraise=True,
    )
    def get_data(
        self,
        url: str,
        method: str = "GET",
        output: str = "json",
        sleep: tuple = (6, 3),
        use_auth_proxies: bool = False,
        max_proxy_delay: timedelta = timedelta(minutes=10),
        ignored_status_codes: Sequence[int] = (),
        **kwargs,
    ):
        """
        Fetches data from a given URL using HTTP requests, with support for proxy configuration, retries, and flexible output formats.

        Args:
            url (str): The target URL to fetch data from.
            method (str, optional): HTTP method to use (e.g., 'GET', 'POST'). Defaults to 'GET'.
            output (str, optional): Output format ('json', 'text', 'soup', 'response'). Defaults to 'json'.
            sleep (tuple, optional): Tuple specifying max and min sleep times (seconds) after request. Defaults to (6, 3).
            use_auth_proxies (bool, optional): Whether to use authenticated proxies. Defaults to False.
            max_proxy_delay (timedelta, optional): Maximum delay for proxy retry logic. Defaults to 10 minutes.
            ignored_status_codes (Sequence[int], optional): Status codes to ignore and return response for. Defaults to ().
            **kwargs: Additional arguments passed to the rnet request method. Supported parameters:
                - timeout (int): Total timeout in seconds (default: 30)
                - read_timeout (int): Read timeout in seconds
                - headers (Dict[str, str]): Custom headers
                - cookies (Dict[str, str]): Custom cookies
                - json (Dict[str, Any]): JSON body for POST/PUT requests
                - form (Dict[str, str] or List[Tuple[str, str]]): Form data
                - body (str or bytes): Raw request body
                - query (List[Tuple[str, str]]): Query parameters
                - auth (str): Authorization header value
                - basic_auth (Tuple[str, Optional[str]]): Basic auth credentials
                - bearer_auth (str): Bearer token
                - allow_redirects (bool): Allow redirects (default: True)
                - max_redirects (int): Max number of redirects
                - And other rnet Request parameters

        Returns:
            Depends on the 'output' argument:
                - 'json': Parsed JSON response.
                - 'text': Response text.
                - 'soup': BeautifulSoup-parsed HTML.
                - 'response': Raw requests.Response object.

        Raises:
            IgnoredHTTPError: If a response status code is in `ignored_status_codes`.
            NotFoundError: If a 404 or 410 status code is returned and not in `ignored_status_codes`.
            BadRequestError: If a 400 status code is returned and not in `ignored_status_codes`.
            EmptyResponseError: If the response has no content.
            ProxyError: On proxy-related errors.
            requests.HTTPError: For other HTTP errors if not ignored.
        """
        # Set default timeout if not provided
        params = kwargs.copy()

        if "timeout" not in params and "read_timeout" not in params:
            params["timeout"] = 30

        r = self._request_with_proxy_retry.retry_with(stop=stop_after_delay(max_proxy_delay))(
            self, url, method, use_auth_proxies, **params
        )

        ban_sleep(*sleep)

        # Get status code as integer
        status_code = r.status_code

        if status_code in ignored_status_codes:
            raise IgnoredHTTPError(message=f"Status {status_code} in ignored_status_codes for URL {url}", response=r)

        error_handlers = {
            404: lambda: NotFoundError(message=f"404 Not Found error for {url}", response=r),
            410: lambda: NotFoundError(message=f"410 Gone error for {url}", response=r),
            400: lambda: BadRequestError(message=f"400 Bad Request error for {url}", response=r),
        }

        if status_code in error_handlers:
            raise error_handlers[status_code]()

        # Use raise_for_status() which raises HTTPError with response accessible via e.response
        r.raise_for_status()

        response_content = r.content
        if not response_content:
            raise EmptyResponseError(message=f"Empty response received from {url} (status {status_code})", response=r)

        r.encoding = "utf-8"
        output_format = {
            "json": lambda: r.json(),
            "text": lambda: r.text,
            "soup": lambda: BeautifulSoup(response_content, "html.parser"),
            "response": lambda: r,
        }

        if output in output_format:
            return output_format[output]()

        raise ValueError(f"Unsupported output format: {output}")



########################################################################################################################
# IMPORTS

import logging
from ftplib import FTP, FTP_TLS
from pathlib import Path
from typing import Any, Dict, List, Optional

########################################################################################################################
# CLASSES

logger = logging.getLogger(__name__)


class FTPInterface:
    def __init__(self, config):
        self.profiles: List[Dict[str, Any]] = []
        self.config = config
        for section in getattr(self.config, "sections", lambda: [])():
            if section.startswith("ftp:"):
                profile_name = section.split(":", 1)[1]
                ftps = self.config[section]["ftps"].lower() == "true"
                ftp_conn = FTP_TLS(self.config[section]["server"]) if ftps else FTP(self.config[section]["server"])  # noqa: S321
                ftp_conn.login(self.config[section]["username"], self.config[section]["password"])
                self.profiles.append({"profile": profile_name, "session": ftp_conn})

        if not self.profiles:
            logger.warning("no ftp section in config")

        self.current_profile: Optional[Dict[str, Any]] = self.profiles[0] if self.profiles else None
        self.ftp = self.current_profile["session"] if self.current_profile else None

    def switch_profile(self, profile_name: str) -> None:
        for profile in self.profiles:
            if profile["profile"] == profile_name:
                self.current_profile = profile
                self.ftp = profile["session"]
                return
        logger.warning(f"Profile {profile_name} not found")

    def upload_file(self, local_file, remote_folder, remote_file=None):
        if not remote_file:
            remote_file = Path(local_file).name

        self._create_remote_dir_tree(full_path=f"/{remote_folder}{remote_file}")

        self.ftp.cwd(f"/{remote_folder}")

        with open(local_file, "rb") as f:
            self.ftp.storbinary(f"STOR {remote_file}", f)

    def download_file(self, local_file, remote_file):
        with open(local_file, "wb") as f:
            self.ftp.retrbinary(f"RETR {remote_file}", f.write)

    def _create_remote_dir_tree(self, full_path):
        dir_tree = full_path.split("/")[0:-1]  # Exclude filename

        for part in dir_tree:
            if not part:
                continue

            try:
                self.ftp.cwd(part)
            except Exception as e:
                logger.warning(f"Error while creating remote directory: {e}")
                self.ftp.mkd(part)
                self.ftp.cwd(part)
